unit rkColorUtil;

// rkColorUtil by Roy Magne Klever
//  2011 by Roy Magne Klever. All rights reserved
//
// This file is not distributable without permission by Roy Magne Klever
// WEB: www.rmklever.com
// Mail: roymagne@rmklever.com
//
// September 2011
//
// You are free to use this in any compiled source projects


interface

uses
  math;

type
  PRGB24 = ^TRGB24;
  TRGB24 = record B, G, R: Byte;
  end;
  PRGBArray = ^TRGBArray;
  TRGBArray = array[0..0] of TRGB24;
  TiRGB = packed record
    B: Integer;
    G: Integer;
    R: Integer;
  end;
  THistogram = array[0..255] of Byte;

  function  RGBtoYi(c: TRGB24): byte;
  procedure RGBtoYCbCr(r,g,b: integer; var y,cb,cr : byte);
  procedure YCbCrtoRGB(y,cb,cr : integer; var r,g,b : byte);
  procedure iRGBtoYCbCr(rgb: TRGB24; var y,cb,cr : byte);
  function  iYCbCrtoRGB(y,cb,cr : integer): TRGB24;
  function iRGBtoS(c: TRGB24): byte;
  procedure RGBtoHSB (const cRed, cGreen, cBlue: byte; var H,S,B: word);
  procedure HSBtoRGB (const H,S,B: word; var cRed, cGreen, cBlue : byte);

const
  ScaleBits   = 16;
  CbCr_Offset = longint(128 shl ScaleBits);
  One_Half    = longint(1) shl (ScaleBits-1);
  R_Y_OFF     = 0;                 { offset to R => Y section }
  G_Y_OFF     = 256;               { offset to G => Y section }
  B_Y_OFF     = 512;               { etc. }
  R_CB_OFF    = 768;
  G_CB_OFF    = 1024;
  B_CB_OFF    = 1280;
  R_CR_OFF    = B_CB_OFF;          { B=>Cb, R=>Cr are the same }
  G_CR_OFF    = 1536;
  B_CR_OFF    = 1792;
  FIX_0_29900 = longint(Round (0.29900 * (1 shl ScaleBits)));
  FIX_0_58700 = longint(Round (0.58700 * (1 shl ScaleBits)));
  FIX_0_11400 = longint(Round (0.11400 * (1 shl ScaleBits)));
  FIX_0_16874 = longint(Round (0.16874 * (1 shl ScaleBits)));
  FIX_0_33126 = longint(Round (0.33126 * (1 shl ScaleBits)));
  FIX_0_50000 = longint(Round (0.50000 * (1 shl ScaleBits)));
  FIX_0_41869 = longint(Round (0.41869 * (1 shl ScaleBits)));
  FIX_0_08131 = longint(Round (0.08131 * (1 shl ScaleBits)));
  FIX_1_40200 = longint(Round( 1.40200 * (1 shl ScaleBits)));
  FIX_1_77200 = longint(Round( 1.77200 * (1 shl ScaleBits)));
  FIX_0_71414 = longint(Round( 0.71414 * (1 shl ScaleBits)));
  FIX_0_34414 = longint(Round( 0.34414 * (1 shl ScaleBits)));
var
  ConvTab  : array[0..2048] of longint;
  Cr_r_tab : array[0..255]  of integer;	{ => table for Cr to R conversion }
  Cb_b_tab : array[0..255]  of integer;	{ => table for Cb to B conversion }
  Cr_g_tab : array[0..255]  of longint;	{ => table for Cr to G conversion }
  Cb_g_tab : array[0..255]  of longint;	{ => table for Cb to G conversion }
  Limiter  : array[-255..512] of integer;

implementation

procedure InitYCbCr;
var
  i : integer;
begin
  for i := 0 to 255 do
  begin
    ConvTab[i+R_Y_OFF]  := FIX_0_29900  * i;
    ConvTab[i+G_Y_OFF]  := FIX_0_58700  * i;
    ConvTab[i+B_Y_OFF]  := FIX_0_11400  * i + One_Half;
    ConvTab[i+R_CB_OFF] := -FIX_0_16874 * i;
    ConvTab[i+G_CB_OFF] := -FIX_0_33126 * i;
    ConvTab[i+B_CB_OFF] := FIX_0_50000  * i + CbCr_Offset + One_Half-1;
    ConvTab[i+G_CR_OFF] := -FIX_0_41869 * i;
    ConvTab[i+B_CR_OFF] := -FIX_0_08131 * i;
  end;
end;

procedure RGBtoY(r,g,b: integer; var y: byte);
begin
  y  := ((ConvTab[r+R_Y_OFF] +ConvTab[g+G_Y_OFF] +ConvTab[b+B_Y_OFF]) shr ScaleBits); // y
end;

function RGBtoYi(c: tRGB24): byte;
begin
  Result:= ((ConvTab[c.R+R_Y_OFF] +ConvTab[c.G+G_Y_OFF] +ConvTab[c.B+B_Y_OFF]) shr ScaleBits); // y
end;

procedure RGBtoYCbCr(r,g,b: integer; var y,cb,cr : byte);
begin
  y  := ((ConvTab[r+R_Y_OFF] +ConvTab[g+G_Y_OFF] +ConvTab[b+B_Y_OFF]) shr ScaleBits); // y
  cb := ((ConvTab[r+R_CB_OFF]+ConvTab[g+G_CB_OFF]+ConvTab[b+B_CB_OFF])shr ScaleBits); // Cb
  cr := ((ConvTab[r+R_CR_OFF]+ConvTab[g+G_CR_OFF]+ConvTab[b+B_CR_OFF])shr ScaleBits); // Cr
end;

procedure iRGBtoYCbCr(rgb: tRGB24; var y,cb,cr : byte);
begin
  y  := ((ConvTab[rgb.r+R_Y_OFF] +ConvTab[rgb.g+G_Y_OFF] +ConvTab[rgb.b+B_Y_OFF]) shr ScaleBits); // y
  cb := ((ConvTab[rgb.r+R_CB_OFF]+ConvTab[rgb.g+G_CB_OFF]+ConvTab[rgb.b+B_CB_OFF])shr ScaleBits); // Cb
  cr := ((ConvTab[rgb.r+R_CR_OFF]+ConvTab[rgb.g+G_CR_OFF]+ConvTab[rgb.b+B_CR_OFF])shr ScaleBits); // Cr
end;

procedure InitRGB;
var
  x, i : integer;
  shift : longint;
begin
  x := -128;
  for i := 0 to 255 do
  begin
    shift := FIX_1_40200 * x + One_Half; {Cr=>R value is nearest int to 1.40200 * x}
    if shift < 0 then Cr_r_tab[i] := (shift shr ScaleBits)or((not longint(0))shl(32-ScaleBits))
    else Cr_r_tab[i] := shift shr ScaleBits;
    shift := FIX_1_77200 * x + One_Half; {Cb=>B value is nearest int to 1.77200 * x}
    if shift < 0 then Cb_b_tab[i] := (shift shr ScaleBits)or((not longint(0))shl(32-ScaleBits))
    else Cb_b_tab[i] := shift shr ScaleBits;
    Cr_g_tab[i] := (-FIX_0_71414) * x;   {Cr=>G value is scaled-up -0.71414 * x}
    Cb_g_tab[i] := (-FIX_0_34414) * x + One_Half;
    Inc(x);
  end;
end;

procedure YCbCrtoRGB(y,cb,cr : integer; var r,g,b : byte);
var
  shift : integer;
begin
  r := limiter[y + Cr_r_tab[cr]];
  shift := Cb_g_tab[cb] + Cr_g_tab[cr];
  if shift < 0 then g := limiter[y+integer((shift shr ScaleBits)or((not longint(0))shl(32-ScaleBits)))]
  else g := limiter[y + shift shr ScaleBits];
  b := limiter[y + Cb_b_tab[cb]];
end;

function iYCbCrtoRGB(y,cb,cr : integer): TRGB24;
var
  shift : integer;
begin
  iYCbCrtoRGB.r := limiter[y + Cr_r_tab[cr]];
  shift := Cb_g_tab[cb] + Cr_g_tab[cr];
  if shift < 0 then iYCbCrtoRGB.g := limiter[y+integer((shift shr ScaleBits)or((not longint(0))shl(32-ScaleBits)))]
  else iYCbCrtoRGB.g := limiter[y + shift shr ScaleBits];
  iYCbCrtoRGB.b := limiter[y + Cb_b_tab[cb]];
end;

procedure MinMax(const i,j,k: byte; var min : integer; var max: word);
begin
  if i > j then begin
    if i > k then max := i else max := k;
    if j < k then min := j else min := k
  end else begin
    if j > k then max := j else max := k;
    if i < k then min := i else min := k
  end;
end;

function iRGBtoS(c: TRGB24): byte;
var
  Delta    : Integer;
  MinValue : Integer;
  tmpH     : Integer;
  H, S, B  : Word;
begin
  tmpH:= 0;
  MinMax(c.R, c.G, c.B, MinValue, B);
  Delta := B - MinValue;
  if B = 0 then S := 0 else S := (255 * Delta) div B;
  if S = 0 then tmpH := 0
  else begin
    if c.R = B then tmpH := (60*(c.G - c.B)) div Delta
      else
    if c.G = B then tmpH := 120 + (60 * (c.B - c.R)) div Delta
      else
    if c.B = B then tmpH := 240 + (60 * (c.R - c.G)) div Delta;
    if tmpH < 0 then tmpH := tmpH + 360;
  end;
  H := tmpH;
  Result:= S;
end;

procedure RGBtoHSB (const cRed, cGreen, cBlue: byte; var H,S,B: word);
var
  Delta    : Integer;
  MinValue : Integer;
  tmpH     : Integer;
begin
  tmpH:= 0;
  MinMax(cRed, cGreen, cBlue, MinValue, B);
  Delta := B - MinValue;
  if B = 0 then S := 0 else S := (255 * Delta) div B;
  if S = 0 then tmpH := 0
  else 
  begin
    if cRed = B then tmpH := (60*(cGreen - cBlue)) div Delta
      else
    if cGreen = B then tmpH := 120 + (60 * (cBlue - cRed)) div Delta
      else
    if cBlue = B then tmpH := 240 + (60 * (cRed - cGreen)) div Delta;
    if tmpH < 0 then tmpH := tmpH + 360;
  end;
  H := tmpH;
end;

procedure HSBtoRGB (const H,S,B: word; var cRed, cGreen, cBlue : byte);
const
  divisor:  Integer = 255*60;
var
  f    :  Integer;
  hTemp:  Integer;
  p,q,t:  Integer;
  VS   :  Integer;
begin
  if s = 0 then
  begin
    cRed:= B;
    cGreen:= B;
    cBlue:= B;
  end
  else
  begin
    if H = 360 then hTemp:= 0 else hTemp:= H;
    f:= hTemp mod 60;
    hTemp:= hTemp div 60;
    VS:= B*S;
    p:= B - VS div 255;
    q:= B - (VS*f) div divisor;
    t:= B - (VS*(60 - f)) div divisor;
    case hTemp of
      0:  begin  cRed := B;   cGreen := t;   cBlue := p  end;
      1:  begin  cRed := q;   cGreen := B;   cBlue := p  end;
      2:  begin  cRed := p;   cGreen := B;   cBlue := t  end;
      3:  begin  cRed := p;   cGreen := q;   cBlue := B  end;
      4:  begin  cRed := t;   cGreen := p;   cBlue := B  end;
      5:  begin  cRed := B;   cGreen := p;   cBlue := q  end;
    end;
  end;
end;

//-- init routine
var
  i : integer;
begin
  initYCbCr;
  initRGB;
  for i := -255 to 512 do begin
    if i < 0 then limiter[i] := 0 else if i > 255 then limiter[i] := 255 else
    limiter[i] := i;
  end;
end.
